/* 
 * This file is part of Pingvo.
 *
 * Copyright (C) 2003-2009 Vad Rulezz <vr5@narod.ru>
 *
 * Pingvo is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <q3scrollview.h>
#include <qsocketnotifier.h>
#include <qtimer.h>
#include <qdom.h>
#include <q3popupmenu.h>
#include <qstatusbar.h>
#include <q3listbox.h>
#include <qthread.h>
#include <qmutex.h>
#include <qwaitcondition.h>
#include <vector>
#include "ui_idxview.h"
#include "ui_artview.h"
#include "ui_confdialog.h"

class IdxViewItem: public Q3ListBoxText {
public:
  char *ittext;
  int refcount;
  IdxViewItem( char *ittext );
  virtual QString text() const;
};

struct lookupEntry {
  short c;
  int word_number, dchilds_count, childs_count;
  int child_offs[0];
};

class Article {
public:
  QString name;

  Article( const QString& name0 ): name(name0) {}
  virtual ~Article() {}
  virtual QString getRichText() = 0;
};

class ArticleXML: public Article {
public:
  ArticleXML( const QString& name, const QDomElement& e0 ):
  	Article(name), e(e0) {}
  virtual ~ArticleXML() {}
  QString getRichText();
protected:
  QDomElement e;
  bool first_m0;

  QString xmlToRichText( QDomElement e );
  QString buildRichRef( QString ref );
};

class ArticleDsl: public Article {
public:
  ArticleDsl( const QString& name, const QString& s0 ):
  	Article(name), s(s0) {}
  QString getRichText();
protected:
  QString s;
};

class Dictionary {
public:
  enum {
  	offline,
	unloading,
  	loading,
	ready
  };

  QString name;
  QString dname, idxname, wfname, treename;
  int menu_idx, state;

  QString cachedText;
  std::vector <lookupEntry*> cachedNodes;

  unsigned char *tree_start;
  int treelen;
  lookupEntry *rootNode;
  std::vector <IdxViewItem*> itMap;

  unsigned char *words_start;
  int wordslen;

  Dictionary();
  virtual ~Dictionary() {}
  lookupEntry *lookupNode( QString word, bool partial_ok, int *match_len = 0 );
  char *itemText( int itnum );

  virtual QString type() = 0;
  virtual Article *getArticle( QString word ) = 0;
protected:
  void getArticle( QByteArray& buf, const QString& word );
};

class DictionaryXML: public Dictionary {
public:
  virtual ~DictionaryXML() {}
  QString type() { return "xml"; }
  Article *getArticle( QString word );
};

class DictionaryDsl: public Dictionary {
public:
  virtual ~DictionaryDsl() {}
  QString type() { return "dsl"; } 
  Article *getArticle( QString word );
};

class LoaderTask {
public:
  enum taskType {
        loadDict,
        unloadDict
  } task;
  int dict_idx;

  LoaderTask( taskType task, int dict_idx );
};

class IdxView;

class DictionaryLoader: public QThread {
public:
  struct loadedItem {
	IdxViewItem *it, *after;
  };

  std::list <LoaderTask*> taskList;
  std::list <loadedItem*> loaded;
  std::list <IdxViewItem*> removed;
  QMutex tlLock, ldLock, rmLock;

  QWaitCondition ack;
  QMutex ackLock;
  bool stopProcessing;

  DictionaryLoader( IdxView *parent );
  virtual void run();
  void loadDictXML( int dict_idx );
  void unloadDictXML( int dict_idx );
  void disposeLookupEntry( lookupEntry *ent );
protected:
  IdxView *idxview;

  void reportEvent( char e );
  void map_words( Dictionary *dict );
  void map_tree( Dictionary *dict );
  void loadDictXML_impl( int dict_idx );
  void unloadDictXML_impl( int dict_idx );
};

class ArticleView;

struct geometryItem {
  int x, y, width, height;
};

class IdxView: public QWidget, public Ui_IdxView {
  Q_OBJECT
  friend class DictionaryLoader;
public:
  IdxView( QWidget *parent = 0, Qt::WFlags f = 0 );
  ~IdxView();
  void initDictXML( Dictionary *ndict, QString fname );
  void loadDictXML( int dictnum );
  void unloadDictXML( int dictnum );
protected:
  int winX, winY, winW, winH;
  int confX, confY, confWidth, confHeight;
  int defArtX, defArtY, defArtWidth, defArtHeight;
  bool xmlMode, configChanged, configured, dont_do_lookup, ign_dict_toggle;
  QTimer *configSaveTimer;
  QSocketNotifier *ev_sn;
  Q3PopupMenu *dictMenu;
  QStatusBar *status_bar;
  std::vector <geometryItem> geometryStack;
  std::vector <ArticleView*> openedBrowsers;
  std::vector <Dictionary*> dictionaries;
  int dict1, dict2, dict3;

  DictionaryLoader dloader;
  lookupEntry *lookupNode( QString text, bool partial_ok, bool longest_match = false );
  lookupEntry *lookupNode( QString text, int& out_dict, bool partial_ok, bool longest_match = false );

  void loadConfig();
  void saveConfig();
  void rebuildDictMenu();
  void updateButtons();
  Article *getArticle( int itDict, QString word );
  QStringList getDictList( QString word );
  ArticleView *findUnpinnedBrowser();
  virtual void keyPressEvent( QKeyEvent *e );
  virtual void moveEvent( QMoveEvent *e );
  virtual void closeEvent( QCloseEvent *e );
  virtual void customEvent( QEvent *e );
public slots:
  void lookupByPrefix();
  void toggleDict1( bool on );
  void toggleDict2( bool on );
  void toggleDict3( bool on );
  void configure();
  void showArticle( Q3ListBoxItem *lbItem );
  void showCurrentArticle();
  void xRecordEventsAvail();
  void confSaveTimeout();
  void articleClosed( ArticleView *av, int x, int y, int w, int h );
  void processAnchorClick( const QString& word, const QString& link );
  void processChangeDict( ArticleView *av, int dict );
  void __processCloseAllViewers();
  void processCloseAllViewers();
  void dictMenuAction( int item_id );
};

class ArticleView: public QWidget, public Ui_ArtView {
  Q_OBJECT
public:
  QString articleName;
  int dictnum;

  ArticleView();
  void loadArticle( Article *art, QStringList ldict );
protected:
  void closeEvent( QCloseEvent *e );
  void keyPressEvent( QKeyEvent *e );
protected slots:
  void dictNumCBChanged( int dict );
  void nextDictRequested();
  void prevDictRequested();
signals:
  void geometryOnClose( ArticleView *av, int x, int y, int w, int h );
  void changeDictRequest( ArticleView *av, int dictNum );
  void closeAllViewersRequest();
};

class confDialog: public QDialog, public Ui_confDialog {
  Q_OBJECT
public:
  int confX, confY, confWidth, confHeight;

  confDialog( QWidget *parent ): QDialog( parent ) {
  	setupUi( this );
  }
protected:
  void closeEvent( QCloseEvent *e );
protected slots:
  void addDictionary();
  void removeDictionary();
  void moveEntryUp();
  void moveEntryDown();
  void accept();
  void reject();
};
