#include "stdafx.h"

#ifdef __linux__
#include "SimpleMemoryMappedFile_linux.hpp"
#else
#include "SimpleMemoryMappedFile.hpp"
#endif

#include "SimpleStdioFile.hpp"
#include "SimpleFile.hpp"
#include "SoundArchive.h"

#ifdef __linux__
#include <sys/ioctl.h>
#include <sys/soundcard.h>

typedef bool BOOL;
typedef int HANDLE;
typedef uint32_t DWORD;
typedef void* LPVOID;

#define APIENTRY
#define TRUE true
#define _T(a) (a)
#define ATLTRACE2(...)
#define ATLASSERT(b)

long int InterlockedIncrement( long int *a ) { return ++*a; }
long int InterlockedDecrement( long int *a ) { return --*a; }
#endif

using std::wstring;
using std::vector;
#pragma comment( lib, "Winmm.lib")


BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    return TRUE;
}


struct SAEntry
{
    DWORD posBeg;
    DWORD posEnd;
    SAEntry() : posBeg( 0), posEnd( 0) {}
    SAEntry( DWORD beg, DWORD end) : posBeg( beg), posEnd( end) {}
};

struct SAItem
{
    DWORD offs;
    wstring word;
    SAItem() {}
    SAItem( DWORD offs_, const wstring & word_) : offs( offs_), word( word_) {}
};


class CSoundArchive : public ISoundArchive
{
    // IBase implementation
    virtual int STDCALL QueryInterface( const char* ifname, void ** object);
    virtual long STDCALL AddRef( void);
    virtual long STDCALL Release( void);
    virtual int  STDCALL GetLastError( void);

#ifdef JNI_SUPPORT
public:
#endif
    // ISoundArchive implementation
    virtual int  STDCALL Open( LPCTSTR pathName);
    virtual void STDCALL Close();
    virtual bool STDCALL FindWord( const uint16_t * word);
    virtual bool STDCALL SayWord( const uint16_t * word);
    virtual bool STDCALL GetSoundData( const uint16_t * word, vector< short> & buffer);
    virtual IWordList* SelectWords( const uint16_t * expression);

public:
    CSoundArchive();
    ~CSoundArchive();

private:
    friend class CWordList;

    long m_cRef;
    int  m_nLastError;
    std::map< std::wstring, SAEntry> m_map;
    CSimpleStdioFileImpl m_file;
    OggVorbis_File m_vf;

    //       .
    static long m_nHeaderSize; 
    static size_t SaRead( void *ptr, size_t size, size_t nmemb, void *datasource);
    static int SaSeek( void *datasource, ogg_int64_t offset, int whence);
    static int SaClose( void *datasource);
    static long SaTell( void *datasource);
    static bool PlayBuffer( vector<short> v);
    static bool SAItemComparator( const SAItem & elem1, const SAItem & elem2);
};

class CWordList : public IWordList
{
    // IWordList implementation
    virtual bool NextWord( uint16_t* &word);
    virtual void Reset();
    virtual void Release();

public:
    CWordList( CSoundArchive *p, const uint16_t * expression);
    ~CWordList();

private:
    CSoundArchive *m_pSoundArchive;
    const uint16_t *m_sExpression;
    std::map< std::wstring, SAEntry>::iterator m_it;
};

CSoundArchive::CSoundArchive() : m_cRef( 0), m_nLastError( 0)
{
}

CSoundArchive::~CSoundArchive()
{
    Close();
}

int STDCALL CSoundArchive::QueryInterface( const char* ifname, void ** object)
{
    if (!strcmp( ifname, "ISoundArchive"))
    {
        *object = (ISoundArchive*) this;
    }
    else
    {
        ATLTRACE2( "Interface %s not found.\n", ifname);
        *object = NULL;
        return errorNoInterface;
    }

    ((IBase*)(*object))->AddRef();
    return errorNoError;
}

long STDCALL CSoundArchive::AddRef( void)
{
    ATLTRACE2( "AddRef() called. m_cRef = %d\n", m_cRef);
    return (int) ::InterlockedIncrement( &m_cRef);
}

long STDCALL CSoundArchive::Release( void)
{
    ATLTRACE2( "Release() called. m_cRef = %d\n", m_cRef);
    if (::InterlockedDecrement( &m_cRef) == 0)
    {
        delete this;
        return 0;
    }

    return m_cRef;
}

int STDCALL CSoundArchive::GetLastError( void)
{
    return m_nLastError;
}

int STDCALL CSoundArchive::Open( LPCTSTR pathName)
{
    ATLASSERT( !m_file);
    ATLASSERT( pathName != NULL);
    ATLTRACE2( "Trying to open %S\n", pathName);

    vector<SAItem> items;
    CSimpleMemoryMappedFile mmfile;
    if (!mmfile.Open( pathName, _T("rb")))
    {
        ATLTRACE2( "Failed to open %s\n", pathName);
        return errorNotFound;
    }

    uint16_t header[] = { 'L', '9', 'S', 'A' };
    uint16_t buf[ 4];

    DWORD dwNumofWords;

    if (mmfile.Read( buf, sizeof(buf)) != sizeof(buf))
        goto clean;

    if (memcmp( header, buf, sizeof(buf)) != 0)
    {
        ATLTRACE2( "Invalid signature.\n");
        goto clean;
    }

    //   .
    mmfile.SeekCur( 1);

    //     .
    dwNumofWords = mmfile.ReadDWord();

    do
    {
        wstring fileName;

        for (;;)
        {
            wchar_t ch = mmfile.ReadWChar();

            if (ch == L'\r')
            {
                //  '\a'  .
                mmfile.SeekCur( 2);
                break;
            }

            fileName += ch;
        }

        //      ( 0xff).
        mmfile.SeekCur( 1);

        //    .
        items.push_back( SAItem( mmfile.ReadDWord(), fileName));

        //   5  .
        mmfile.SeekCur( 5);
    }
    while (items.size() != dwNumofWords);

    m_nHeaderSize = mmfile.Tell();

    ATLTRACE2( "Found %u words\n", items.size());

    //    .
    std::sort( items.begin(), items.end(), SAItemComparator);

    ATLTRACE2( "Sorted\n");

    ov_callbacks OvCallBacks;
    OvCallBacks.read_func  = SaRead;
    OvCallBacks.seek_func  = SaSeek;
    OvCallBacks.close_func = SaClose;
    OvCallBacks.tell_func  = SaTell;

    mmfile.Close();

    if (!m_file.Open( pathName, _T("rb")))
    {
        ATLTRACE2( "Failed to reopen %s\n", pathName);
        Close();
        return errorNotFound;
    }

    ogg_int64_t totalPcmSamples;

    if (ov_open_callbacks( (FILE*) m_file, &m_vf, 0, -1, OvCallBacks) != 0)
        goto clean;

    ATLTRACE2( "Open() succeeded.\n");

    totalPcmSamples = ov_pcm_total( &m_vf, 0);

    //      .
    for (vector<SAItem>::iterator it = items.begin(), last = items.end() - 1; it != items.end(); ++it)
    {
        SAEntry entry;
        entry.posBeg = it->offs;
        entry.posEnd = (it == last) ? (DWORD) totalPcmSamples : (it + 1)->offs;
        m_map[ it->word] = entry;
    }

    ATLTRACE2( "map filled\n");

    return errorNoError;

clean:
    Close();
    return errorFail;
}

void STDCALL CSoundArchive::Close()
{
    ATLTRACE2( "Close() called\n");
    m_map.clear();
    ov_clear( &m_vf);
    m_file.Close();
}

bool STDCALL CSoundArchive::FindWord( const uint16_t * word0)
{
    wstring word;
    while ( *word0 )  word += *word0++;

    ATLTRACE2( L"Trying to find word %s\n", word);
    std::map< std::wstring, SAEntry>::iterator it = m_map.find( word);
    return (it != m_map.end());
}

bool STDCALL CSoundArchive::SayWord( const uint16_t * word)
{
    vector< short> buffer;

    if (!GetSoundData( word, buffer))
        return true;

    PlayBuffer( buffer);
    return true;
}

bool STDCALL CSoundArchive::GetSoundData( const uint16_t * word0, vector< short> & buffer)
{
    wstring word;
    while ( *word0 ) word += *word0++;

    ATLASSERT( buffer.empty());
    ATLTRACE2( "GetSoundData() called.\n");

    std::map< std::wstring, SAEntry>::iterator it = m_map.find( word);
    if (it == m_map.end())
        return false;

    if (ov_pcm_seek( &m_vf, it->second.posBeg) != 0)
        return false;

    int bitstream;
    buffer.reserve( 0x100000); // 2MB

    for (size_t idx = 0;;)
    {
        buffer.resize( idx + 2048);  

        long cbRead = ov_read( &m_vf, (char*) &buffer[ idx], 4096, 0, 2, 1, &bitstream);
        if (cbRead <= 0)
            break;

        DWORD curPos = (DWORD) ov_pcm_tell( &m_vf);
        if (curPos >= it->second.posEnd)
            break;

        idx += cbRead / 2;
    }

    return true;
}

IWordList* CSoundArchive::SelectWords( const uint16_t * expression)
{
    ATLASSERT( expression != NULL);
    return new CWordList( this, expression);
}

#ifdef __linux__
extern "C" {
IBase* CreateInstance()
#else
extern "C" __declspec( dllexport) IBase* CreateInstance()
#endif
{
    IBase *pi = (ISoundArchive*) new CSoundArchive;
    pi->AddRef();
    return pi;
}
#ifdef __linux__
}
#endif

size_t CSoundArchive::SaRead( void *ptr, size_t size, size_t nmemb, void *datasource)
{
    return fread( ptr, size, nmemb, (FILE*) datasource);
}

int CSoundArchive::SaSeek( void *datasource, ogg_int64_t offset, int whence)
{
    return fseek((FILE*) datasource, (long) offset + m_nHeaderSize, whence);
}

int CSoundArchive::SaClose( void *datasource)
{
    return 0;
}

long CSoundArchive::SaTell( void *datasource)
{
    return ftell( (FILE*) datasource);
}

#ifdef __linux__
bool CSoundArchive::PlayBuffer( vector<short> v ) {
  int fd, val;

  fd = open( "/dev/dsp", O_WRONLY );
  if ( !fd ) return false;

  val = 0;
  ioctl( fd, SNDCTL_DSP_STEREO, &val );
  val = 48000;
  ioctl( fd, SNDCTL_DSP_SPEED, &val );
  val = AFMT_S16_LE;
  ioctl( fd, SNDCTL_DSP_SETFMT, &val );

  write( fd, &v[0], v.size() * sizeof(short) );

  close( fd );

  return true;
}
#else
bool CSoundArchive::PlayBuffer( vector<short> v)
{
    MMRESULT result;
    WAVEFORMATEX fmt;
    HWAVEOUT phwo;

    ZeroMemory( &fmt, sizeof(fmt));
    fmt.wFormatTag = WAVE_FORMAT_PCM;
    fmt.nChannels = 1;
    fmt.nSamplesPerSec = 48000;
    fmt.wBitsPerSample = 16;
    fmt.nBlockAlign = (fmt.nChannels * fmt.wBitsPerSample) / 8;
    fmt.nAvgBytesPerSec = fmt.nSamplesPerSec * fmt.nBlockAlign;

    result = waveOutOpen( &phwo, WAVE_MAPPER, &fmt, 0, 0, CALLBACK_NULL);
    if (result != MMSYSERR_NOERROR)
        return false;

    waveOutSetVolume( phwo, 0xFFFFFFFF);

    WAVEHDR hdr;
    ZeroMemory( &hdr, sizeof(hdr));
    hdr.lpData = (LPSTR) &v[0];
    hdr.dwBufferLength = (DWORD) (v.size() * sizeof(short));

    result = waveOutPrepareHeader( phwo, &hdr, sizeof(hdr));
    if (result != MMSYSERR_NOERROR)
    {
        waveOutClose( phwo);
        return false;
    }

    result = waveOutWrite( phwo, &hdr, sizeof(hdr));
    if (result != MMSYSERR_NOERROR)
    {
        waveOutClose( phwo);
        return false;
    }

    //   .
    do { Sleep( 1); } while (!(hdr.dwFlags & WHDR_DONE));
    
    waveOutUnprepareHeader( phwo, &hdr, sizeof(hdr));
    waveOutClose( phwo);
    return true;
}
#endif

bool CSoundArchive::SAItemComparator( const SAItem & elem1, const SAItem & elem2)
{
    return elem1.offs < elem2.offs;
}

long CSoundArchive::m_nHeaderSize = 0;

///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

CWordList::CWordList( CSoundArchive *p, const uint16_t * expression) : \
    m_pSoundArchive( p), m_sExpression( expression)
{
    ATLASSERT( m_sExpression);
    Reset();
}

CWordList::~CWordList()
{
}

bool CWordList::NextWord( uint16_t* &word)
{
    wstring word0;
    int i;

    if (m_it != m_pSoundArchive->m_map.end())
    {
	word0 = m_it->first;
	word = (uint16_t*) malloc( (word0.length() + 1) * sizeof(uint16_t) );

	for ( i = 0; word0[i]; i++ ) word[i] = word0[i];
	word[i] = 0;

        ++m_it;
        return true;
    }
    else
        return false;
}

void CWordList::Reset()
{
    ATLASSERT( _CrtIsValidPointer( m_pSoundArchive, sizeof(CSoundArchive), FALSE));
    m_it = m_pSoundArchive->m_map.begin();
}

void CWordList::Release()
{
    // ,     CSoundArchive
    delete this;
}

#ifdef JNI_SUPPORT
#include <jni.h>

extern "C" {
JNIEXPORT jint JNICALL Java_vr5_SoundArchive_ctor( JNIEnv *env, jobject obj ) {
  return (jint) new CSoundArchive;
}

JNIEXPORT void JNICALL Java_vr5_SoundArchive_dtor( JNIEnv *env, jobject obj,
	jint ref ) {
  delete (CSoundArchive*)ref;
}

JNIEXPORT jint JNICALL Java_vr5_SoundArchive_Open( JNIEnv *env, jobject obj,
	jint ref, jstring pathName ) {
  jboolean b;
  return ((CSoundArchive*)ref)->Open( env->GetStringUTFChars( pathName, &b ) );
}

JNIEXPORT void JNICALL Java_vr5_SoundArchive_Close( JNIEnv *env, jobject obj,
	jint ref ) {
  ((CSoundArchive*)ref)->Close();
}

JNIEXPORT jint JNICALL Java_vr5_SoundArchive_SayWord( JNIEnv *env, jobject obj,
	jint ref, jstring word ) {
  jboolean b;
  return ((CSoundArchive*)ref)->SayWord( env->GetStringChars( word, &b ) );
}
} // extern
#endif
